# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""A wrapper around `timm.create_model() <https://rwightman.github.io/pytorch-image-models/#load-a-pretrained-model>`_
used to create :class:`.ComposerClassifier`."""

from typing import Optional

from composer.models.tasks import ComposerClassifier
from composer.utils.import_helpers import MissingConditionalImportError

__all__ = ['composer_timm']


def composer_timm(model_name: str,
                  pretrained: bool = False,
                  num_classes: int = 1000,
                  drop_rate: float = 0.0,
                  drop_path_rate: Optional[float] = None,
                  drop_block_rate: Optional[float] = None,
                  global_pool: Optional[str] = None,
                  bn_momentum: Optional[float] = None,
                  bn_eps: Optional[float] = None) -> ComposerClassifier:
    """A wrapper around `timm.create_model() <https://rwightman.github.io/pytorch-image-models/#load-a-pretrained-
    model>`_ used to create :class:`.ComposerClassifier`.

    Args:
        model_name (str): timm model name e.g: ``"resnet50"``. List of models can be found at
            `PyTorch Image Models <https://github.com/rwightman/pytorch-image-models>`_.
        pretrained (bool, optional): Imagenet pretrained. Default: ``False``.
        num_classes (int, optional): The number of classes. Needed for classification tasks. Default: ``1000``.
        drop_rate (float, optional): Dropout rate. Default: ``0.0``.
        drop_path_rate (float, optional): Drop path rate (model default if ``None``). Default: ``None``.
        drop_block_rate (float, optional): Drop block rate (model default if ``None``). Default: ``None``.
        global_pool (str, optional): Global pool type, one of (``"fast"``, ``"avg"``, ``"max"``, ``"avgmax"``, ``"avgmaxc"``). Model default if ``None``. Default: ``None``.
        bn_momentum (float, optional): BatchNorm momentum override (model default if ``None``). Default: ``None``.
        bn_eps (float, optional): BatchNorm epsilon override (model default if ``None``). Default: ``None``.

    Returns:
        ComposerModel: instance of :class:`.ComposerClassifier` with the specified TIMM model.

    Resnet18 Example:

    .. testcode::

        from composer.models import composer_timm

        model = composer_timm(model_name='resnet18')  # creates a timm resnet18
    """
    try:
        import timm
    except ImportError as e:
        raise MissingConditionalImportError(extra_deps_group='timm', conda_package='timm>=0.5.4',
                                            conda_channel=None) from e
    model = timm.create_model(  # type: ignore (third-party)
        model_name=model_name,
        pretrained=pretrained,
        num_classes=num_classes,
        drop_rate=drop_rate,
        drop_path_rate=drop_path_rate,
        drop_block_rate=drop_block_rate,
        global_pool=global_pool,
        bn_momentum=bn_momentum,
        bn_eps=bn_eps)

    composer_model = ComposerClassifier(module=model)
    return composer_model



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Log to `Tensorboard <https://www.tensorflow.org/tensorboard/>`_."""

from pathlib import Path
from typing import Any, Dict, Optional

from composer.core.state import State
from composer.loggers.logger import Logger, format_log_data_value
from composer.loggers.logger_destination import LoggerDestination
from composer.utils import MissingConditionalImportError, dist

__all__ = ['TensorboardLogger']


class TensorboardLogger(LoggerDestination):
    """Log to `Tensorboard <https://www.tensorflow.org/tensorboard/>`_.

    If you are accessing your logs from a cloud bucket, like S3, they will be
    in `{your_bucket_name}/tensorboard_logs/{run_name}` with names like
    `events.out.tfevents-{run_name}-{rank}`.

    If you are accessing your logs locally (from wherever you are running composer), the logs
    will be in the relative path: `tensorboard_logs/{run_name}` with names starting with
    `events.out.tfevents.*`

    Args:
        log_dir (str, optional): The path to the directory where all the tensorboard logs
            will be saved. This is also the value that should be specified when starting
            a tensorboard server. e.g. `tensorboard --logdir={log_dir}`. If not specified
            `./tensorboard_logs` will be used.
        flush_interval (int, optional): How frequently by batch to flush the log to a file.
            For example, a flush interval of 10 means the log will be flushed to a file
            every 10 batches. The logs will also be automatically flushed at the start and
            end of every evaluation phase (`EVENT.EVAL_START` and `EVENT.EVAL_END` ),
            the end of every epoch (`EVENT.EPOCH_END`), and the end of training
            (`EVENT.FIT_END`). Default: ``100``.
        rank_zero_only (bool, optional): Whether to log only on the rank-zero process.
            Recommended to be true since the rank 0 will have access to most global metrics.
            A setting of `False` may lead to logging of duplicate values.
            Default: :attr:`True`.
    """

    def __init__(self, log_dir: Optional[str] = None, flush_interval: int = 100, rank_zero_only: bool = True):
        try:
            from torch.utils.tensorboard import SummaryWriter
        except ImportError as e:
            raise MissingConditionalImportError(extra_deps_group='tensorboard',
                                                conda_package='tensorboard',
                                                conda_channel='conda-forge') from e

        self.log_dir = log_dir
        self.flush_interval = flush_interval
        self.rank_zero_only = rank_zero_only
        self.writer: Optional[SummaryWriter] = None
        self.run_name: Optional[str] = None
        self.hyperparameters: Dict[str, Any] = {}
        self.current_metrics: Dict[str, Any] = {}

    def log_hyperparameters(self, hyperparameters: Dict[str, Any]):

        if self.rank_zero_only and dist.get_global_rank() != 0:
            return
        # Lazy logging of hyperparameters b/c Tensorboard requires a metric to pair
        # with hyperparameters.
        formatted_hparams = {
            hparam_name: format_log_data_value(hparam_value) for hparam_name, hparam_value in hyperparameters.items()
        }
        self.hyperparameters.update(formatted_hparams)

    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None):
        if self.rank_zero_only and dist.get_global_rank() != 0:
            return

        # Keep track of most recent metrics to use for `add_hparams` call.
        self.current_metrics.update(metrics)

        for tag, metric in metrics.items():
            if isinstance(metric, str):  # Will error out with weird caffe2 import error.
                continue
            # TODO: handle logging non-(scalars/arrays/tensors/strings)
            # If a non-(scalars/arrays/tensors/strings) is passed, we skip logging it,
            # so that we do not crash the job.
            try:
                assert self.writer is not None
                self.writer.add_scalar(tag, metric, global_step=step)
            # Gets raised if data_point is not a tensor, array, scalar, or string.
            except NotImplementedError:
                pass

    def init(self, state: State, logger: Logger) -> None:
        self.run_name = state.run_name

        # We fix the log_dir, so all runs are co-located.
        if self.log_dir is None:
            self.log_dir = 'tensorboard_logs'

        self._initialize_summary_writer()

    def _initialize_summary_writer(self):
        from torch.utils.tensorboard import SummaryWriter

        assert self.run_name is not None
        assert self.log_dir is not None
        # We name the child directory after the run_name to ensure the run_name shows up
        # in the Tensorboard GUI.
        summary_writer_log_dir = Path(self.log_dir) / self.run_name

        # Disable SummaryWriter's internal flushing to avoid file corruption while
        # file staged for upload to an ObjectStore.
        flush_secs = 365 * 3600 * 24
        self.writer = SummaryWriter(log_dir=summary_writer_log_dir, flush_secs=flush_secs)

    def batch_end(self, state: State, logger: Logger) -> None:
        if int(state.timestamp.batch) % self.flush_interval == 0:
            self._flush(logger)

    def epoch_end(self, state: State, logger: Logger) -> None:
        self._flush(logger)

    def eval_end(self, state: State, logger: Logger) -> None:
        # Give the metrics used for hparams a unique name, so they don't get plotted in the
        # normal metrics plot.
        metrics_for_hparams = {
            'hparams/' + name: metric
            for name, metric in self.current_metrics.items()
            if 'metric' in name or 'loss' in name
        }
        assert self.writer is not None
        self.writer.add_hparams(hparam_dict=self.hyperparameters,
                                metric_dict=metrics_for_hparams,
                                run_name=self.run_name)
        self._flush(logger)

    def fit_end(self, state: State, logger: Logger) -> None:
        self._flush(logger)

    def _flush(self, logger: Logger):
        # To avoid empty files uploaded for each rank.
        if self.rank_zero_only and dist.get_global_rank() != 0:
            return

        if self.writer is None:
            return
        # Skip if no writes occurred since last flush.
        if not self.writer.file_writer:
            return

        self.writer.flush()

        file_path = self.writer.file_writer.event_writer._file_name
        event_file_name = Path(file_path).stem

        logger.upload_file(remote_file_name=('tensorboard_logs/{run_name}/' +
                                             f'{event_file_name}-{dist.get_global_rank()}'),
                           file_path=file_path,
                           overwrite=True)

        # Close writer, which creates new log file.
        self.writer.close()

    def close(self, state: State, logger: Logger) -> None:
        del state  # unused
        self._flush(logger)
        self.writer = None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Optional

import numpy as np

from composer.core import Dataset
from composer.utils import MissingConditionalImportError

try:
    import ffcv
    ffcv_installed = True
except ImportError:
    ffcv_installed = False

log = logging.getLogger(__name__)

__all__ = ['write_ffcv_dataset', 'ffcv_monkey_patches']


def _require_ffcv():
    if not ffcv_installed:
        raise MissingConditionalImportError(extra_deps_group='ffcv', conda_package='ffcv')


def ffcv_monkey_patches():
    _require_ffcv()

    # ffcv's __len__ function is expensive as it always calls self.next_traversal_order which does shuffling.
    # Composer calls len(dataloader) function in training loop for every batch and thus len function causes 2x slowdown.
    # ffcv's __len__ is fixed in 1.0.0 branch but for another reason (https://github.com/libffcv/ffcv/issues/163).
    def new_len(self):
        if not hasattr(self, 'init_traversal_order'):
            self.init_traversal_order = self.next_traversal_order()
        if self.drop_last:
            return len(self.init_traversal_order) // self.batch_size
        else:
            return int(np.ceil(len(self.init_traversal_order) / self.batch_size))

    ffcv.loader.loader.Loader.__len__ = new_len


def write_ffcv_dataset(dataset: Optional[Dataset] = None,
                       write_path: str = '/tmp/dataset.ffcv',
                       max_resolution: Optional[int] = None,
                       num_workers: int = 16,
                       write_mode: str = 'raw',
                       compress_probability: float = 0.50,
                       jpeg_quality: float = 90,
                       chunk_size: int = 100):
    """Converts PyTorch compatible ``dataset`` into FFCV format at filepath ``write_path``.

    Args:
        dataset (Iterable[Sample]): A PyTorch dataset. Default: ``None``.
        write_path (str): Write results to this file. Default: ``"/tmp/dataset.ffcv"``.
        max_resolution (int): Limit resolution if provided. Default: ``None``.
        num_workers (int): Numbers of workers to use. Default: ``16``.
        write_mode (str): Write mode for the dataset. Default: ``'raw'``.
        compress_probability (float): Probability with which image is JPEG-compressed. Default: ``0.5``.
        jpeg_quality (float): Quality to use for jpeg compression. Default: ``90``.
        chunk_size (int): Size of chunks processed by each worker during conversion. Default: ``100``.
    """

    _require_ffcv()
    if dataset is None:
        raise ValueError('dataset should not be None.')

    log.info(f'Writing dataset in FFCV <file>.ffcv format to {write_path}.')
    writer = ffcv.writer.DatasetWriter(write_path, {
        'image':
            ffcv.fields.RGBImageField(write_mode=write_mode,
                                      max_resolution=max_resolution,
                                      compress_probability=compress_probability,
                                      jpeg_quality=jpeg_quality),
        'label':
            ffcv.fields.IntField()
    },
                                       num_workers=num_workers)
    writer.from_indexed_dataset(dataset, chunksize=chunk_size)
